/*
 *  Copyright (C) 2007 Free Software Foundation, Inc.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <getopt.h>
#include <errno.h>

#include "key_tab.h"

#define BUF_SIZE	65536

#define OFS_LEN1	2
#define OFS_LEN2	0x2C
#define OFS_CSUM	5
#define OFS_HKEY	6
#define OFS_TIME	8
#define OFS_MOFS	0xA
#define OFS_MLEN	0xB
#define OFS_REDX	0xC

#define MAX_MLEN	32

#define APP_NAME	"grub-mkrom"

#define valueat(buf,ofs,type)	*((type*)(((char*)&buf)+ofs))

void usage(int status)
{
  fprintf(stderr,
          "Usage:\n"
          "  grub-mkrom [OPTIONS] romboot.img core.img output\n\n"
          "Options:\n"
          "  -m message\t\tBoot message\n"
          "  -t timeout\t\tTimeout in seconds\n"
          "  -k key\t\tHotkey\n"
          "  -x edx\t\tValue of EDX\n");
  exit(status);
}

void quit(int syserr,char *fmt, ...)
{
  va_list argptr;

  va_start(argptr, fmt);

  fprintf(stderr, APP_NAME ": ");
  vfprintf(stderr, fmt, argptr);
  if (syserr)
    fprintf(stderr,": %s\n", strerror(errno));
  else
    fprintf(stderr,"\n");

  va_end(argptr);

  exit(1);
}

char* str_upcase(char* str)
{
  int i;

  for (i=0;str[i];i++)
    if ((str[i]>='a') && (str[i]<='z'))
      str[i]-='a'-'A';

  return str;
}

int get_keycode(char* key)
{
  int idx,i;
  char *str;

  if ((key[0]=='0') && (key[1]=='x'))
    {
      int code;

      code=strtoul(key,NULL,0);
      return ((code<=0) || (code>=0xFFFF))?0:code;
    }
  str=key;
  str_upcase(str);
  idx=0;
  if (! strncmp(str,"SHIFT-",6))
    {
      idx=1;
      str+=6;
    }
  else if (! strncmp(str,"CTRL-",5))
    {
      idx=2;
      str+=5;
    }
  else if (! strncmp(str,"ALT-",4))
    {
      idx=3;
      str+=4;
    }
  for (i=0;key_table[i].name;i++)
    {
      if (! strcmp(str,key_table[i].name))
        return key_table[i].code[idx];
    }
  return 0;
}

unsigned char buf[BUF_SIZE];

int main(int argc, char **argv)
{
  char ch;
  FILE *ff;
  long ofs,len,i;
  int is_pci;
  unsigned char csum;
  char msg[MAX_MLEN];
  int timeout,hotkey;
  long edx;

  msg[0]=0;
  timeout=-1;
  hotkey=-1;
  edx=-1;
  while ((ch=getopt(argc,argv, "hm:k:t:x:"))!= -1)
    switch (ch) {
    case 'm':
      {
        if (strlen(optarg)>=MAX_MLEN)
          quit(0, "Message too long");
        strcpy(msg, optarg);
        break;
      }
    case 't':
      {
        timeout=strtoul(optarg,NULL,0) * 18;
        break;
      }
    case 'k':
      {
        int key;

        key=get_keycode(optarg);
        if (key)
          hotkey=key;
        break;
      }
    case 'x':
      {
        edx=strtoul(optarg,NULL,0);
        break;
      }
    case 'h':
    default:
      usage(0);
    }

  if (optind + 3 > argc)
    usage(1);

  ff=fopen(argv[optind], "rb");
  if (ff == NULL)
    quit(1, "Unable to open %s", argv[optind]);

  fseek(ff,0,SEEK_END);
  len = ftell(ff);
  if (len >= BUF_SIZE)
    {
      fclose(ff);
      quit(0, "%s too long", argv[optind]);
    }

  fseek(ff, 0, SEEK_SET);
  if (fread(buf, len, 1, ff)!=1)
    {
      fclose(ff);
      quit(0, "%s read error", argv[optind]);
    }

  fclose(ff);

  if (len < *((unsigned short*)&buf[0]))
    quit(0, "%s too short", argv[optind]);

  ofs = valueat(buf,0,unsigned short);
  valueat(buf,0,unsigned short) = 0xAA55;
  is_pci=(memcmp(&buf[0x1C],"PCIR",4)==0);

  if (msg[0])
    strcpy(&buf[buf[OFS_MOFS]],msg);

  if (timeout!=-1)
    valueat(buf, OFS_TIME, unsigned short) = timeout;

  if (hotkey!=-1)
    valueat(buf, OFS_HKEY, unsigned short) = hotkey;


  if (edx!=-1)
    valueat(buf, OFS_REDX, unsigned long) = edx;

  ff=fopen(argv[optind+1], "rb");
  if (ff == NULL)
    quit(1, "Unable to open %s", argv[optind+1]);

  fseek(ff,0,SEEK_END);
  len = ftell(ff);
  if (len<=512)
    {
      fclose(ff);
      quit(0, "%s too short", argv[optind+1]);
    }

  if (ofs + len > BUF_SIZE)
    {
      fclose(ff);
      quit(0, "%s %d bytes too long", argv[optind+1], ofs + len - BUF_SIZE);
    }

  fseek(ff, 512, SEEK_SET);
  len -= 512;

  if (fread(&buf[ofs], len, 1, ff)!=1)
    {
      fclose(ff);
      quit(0, "%s read error", argv[optind+1]);
    }

  fclose(ff);

  len = (len + ofs + 511) & (~0x1FF);

  buf[OFS_LEN1] = len >> 9;
  if (is_pci)
    buf[OFS_LEN2] = buf[OFS_LEN1];

  csum = 0;
  for (i=0;i<len;i++)
    csum += buf[i];

  buf[OFS_CSUM] = (unsigned char)(0x100 - (int) csum);

  ff=fopen(argv[optind+2], "wb");
  if (ff == NULL)
    quit(1, "Unable to open %s", argv[optind+2]);

  if (fwrite(buf, len, 1, ff)!=1)
    {
      fclose(ff);
      quit(0, "%s read error", argv[optind+2]);
    }

  fclose(ff);

  return 0;
}
